/*******************************************************************************
*    NetTools 1.1 for PalmOS                                                   *
*    Copyright (C) 2006-2007 Luca Bertoncello                                  *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <PalmOS.h>
#include <PalmTypes.h>
#include <PalmCompatibility.h>
#include <ConnectionMgr.h>
#include <SystemMgr.h>
#include <NetMgr.h>
#include <Unix/sys_socket.h>
#include <netinet_in.h>
#include "netTools.h"
#include "prefs.h"

#define maxServiceNameLen       32
#define appID                   'NetT'
#define versionID               1

typedef struct
{
  Char IP[20];
  Char DNS1[20];
  Char DNS2[20];
  Char Speed[10];
  Char HwName[20];
  Char InB[20];
  Char OutB[20];
  int ifUp;
} connectionDataType;
typedef struct
{
  Err error;
  UInt16 numServices;
  MemHandle IDListH;
  MemHandle nameListH;
} serviceListType;
typedef serviceListType * serviceListPtr;
typedef UInt32 ServiceIDType;
UInt32 * connectionsID;
Char **itemList;
serviceListType serviceList;
int totConnections, connected, connectedToNet;
static FormPtr gpForm;
UInt16 AppNetRefnum, Instance;
UInt32 Creator;
connectionDataType connData;
Preferences gAppPrefs;

UInt32 GetActiveService()
{
  UInt16 cardNo;
  LocalID ID;
  DmSearchStateType theSearchState;
  UInt32 dwResult;
  UInt32 currServiceID;

  DmGetNextDatabaseByTypeCreator(true, &theSearchState, sysFileTPanel, sysFileCNetworkPanel,true, &cardNo, &ID);
  SysAppLaunch(cardNo, ID, 0/*launchFlags*/, sysSvcLaunchCmdGetServiceID, (MemPtr) &currServiceID, &dwResult); 

  return currServiceID;
}

void SetActiveService(UInt32 currServiceID)
{
  UInt16 cardNo;
  LocalID ID;
  DmSearchStateType theSearchState;
  UInt32 dwResult;

  DmGetNextDatabaseByTypeCreator(true, &theSearchState, sysFileTPanel, sysFileCNetworkPanel,true, &cardNo, &ID);
  SysAppLaunch(cardNo, ID, 0/*launchFlags*/, sysSvcLaunchCmdSetServiceID, &currServiceID, &dwResult); 
}

void getConnectionData()
{
  Err err = 0;
  Char an_addr[32];
  UInt16 bufLen;
  UInt32 dns1, dns2;
  NetMasterPBType pb;

  if(connected == 1)
  {
    dns1 = dns2 = 0;
    StrCopy(connData.DNS1, "");
    StrCopy(connData.DNS2, "");
    NetLibSettingGet(AppNetRefnum, netSettingRTPrimaryDNS, &dns1, &bufLen);
    NetLibSettingGet(AppNetRefnum, netSettingRTSecondaryDNS, &dns2, &bufLen);
    StrPrintF(connData.DNS1, "%s", NetLibAddrINToA(AppNetRefnum, dns1, an_addr));
    if(StrCompare(connData.DNS1, "0.0.0.0") == 0)
    {
      NetLibSettingGet(AppNetRefnum, netSettingRTPrimaryDNS, &dns1, &bufLen);
      StrPrintF(connData.DNS1, "%s", NetLibAddrINToA(AppNetRefnum, dns1, an_addr));
    }
    if(dns1 == dns2)
      StrPrintF(connData.DNS2, "");
    else
      StrPrintF(connData.DNS2, "%s", NetLibAddrINToA(AppNetRefnum, dns2, an_addr));
    for(pb.param.interfaceInfo.index = 0; 1; pb.param.interfaceInfo.index++)
    {
      err = NetLibMaster(AppNetRefnum, netMasterInterfaceInfo, &pb, sysTicksPerSecond * 2);
      if(err == netErrParamErr || pb.param.interfaceInfo.index == 5)
        break;
      if(err)
        break;
      else
      {
        if(StrCompare(pb.param.interfaceInfo.drvrName, "loop") != 0)
        {
          Creator = pb.param.interfaceInfo.creator;
          StrPrintF(connData.IP, "%s", NetLibAddrINToA(AppNetRefnum, pb.param.interfaceInfo.ipAddr, an_addr));
          connData.ifUp = pb.param.interfaceInfo.ifUp;
          StrPrintF(connData.Speed, "%ld", pb.param.interfaceInfo.speed);
          StrPrintF(connData.HwName, "%s", pb.param.interfaceInfo.hwName);
          pb.param.interfaceStats.index = pb.param.interfaceInfo.index;
          NetLibMaster(AppNetRefnum, netMasterInterfaceStats, &pb, sysTicksPerSecond*2);
          StrPrintF(connData.InB, "%ld", pb.param.interfaceStats.inOctets);
          StrPrintF(connData.OutB, "%ld", pb.param.interfaceStats.outOctets);
        }
      }
    }
  }
  else
    StrCopy(connData.IP, "0.0.0.0");
}

Err connectToService()
{
  if(connected == 0)
  {
    NetLibOpen(AppNetRefnum, &Instance);
    connected = 1;
  }

  return NetLibIFUp(AppNetRefnum, Creator, Instance);
}

void disconnectToService()
{
  NetLibIFDown(AppNetRefnum, Creator, Instance, 0);
}

void initConnectionsList()
{
  UInt16 cardNo;
  LocalID ID;
  DmSearchStateType theSearchState;
  UInt32 dwResult;
  Char *p, serviceName[maxServiceNameLen];
  int i = 0, j = 0, k = 0;
  FormType* frmP;
  ListType *lstP;

  frmP = FrmGetActiveForm();
  lstP = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConnectionsList));
  MemSet(&serviceList, sizeof(serviceListType), 0);
  DmGetNextDatabaseByTypeCreator(true, &theSearchState, sysFileTPanel, sysFileCNetworkPanel,true, &cardNo, &ID);
  SysAppLaunch(cardNo, ID, 0/*launchFlags*/, sysSvcLaunchCmdGetServiceList, (MemPtr) &serviceList, &dwResult);
  p = MemHandleLock(serviceList.nameListH);
  connectionsID = MemHandleLock(serviceList.IDListH);

  totConnections = serviceList.numServices;
  itemList = (Char **) MemPtrNew(serviceList.numServices * sizeof(Char *));
  while(i < serviceList.numServices)
  { 
    while(p[j] != '\0')
    {
      if(p[j] == '\0')
        j++; //escape the null byte.
      else 
        serviceName[k++] = p[j++];
    }
    serviceName[k] = '\0'; //terminate the string.
    itemList[i] = (char *) MemPtrNew(StrLen(serviceName) + 1);
    StrCopy(itemList[i], serviceName);
    j++;
    i++;
    k = 0;
  }

  LstSetListChoices(lstP, itemList, serviceList.numServices);

  MemHandleUnlock(serviceList.nameListH);
  MemPtrFree(p);
}

void displayNetInfo()
{
  FormType* frmP;
  ListType *lstP;
  Int16 sel;

  frmP = FrmGetActiveForm();
  lstP = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConnectionsList));

  sel = LstGetSelection(lstP);

  connectedToNet = 1;
  CtlHideControl(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonConnect)));
  CtlShowControl(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonDisconnect)));
  CtlSetEnabled(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonEdit)), false);
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldIP)), connData.IP);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldIP)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS1)), connData.DNS1);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS1)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS2)), connData.DNS2);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS2)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldSpeed)), connData.Speed);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldSpeed)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldHwName)), connData.HwName);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldHwName)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldInB)), connData.InB);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldInB)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldOutB)), connData.OutB);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldOutB)));
}

void hideNetInfo()
{
  FormType* frmP;

  frmP = FrmGetActiveForm();

  connectedToNet = 0;
  CtlHideControl(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonDisconnect)));
  CtlShowControl(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonConnect)));
  CtlSetEnabled(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idButtonEdit)), true);
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldIP)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldIP)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS1)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS1)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS2)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldDNS2)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldSpeed)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldSpeed)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldHwName)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldHwName)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldInB)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldInB)));
  FldSetTextPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldOutB)), NULL);
  FldDrawField(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, idFieldOutB)));
}

void CallPanel(UInt32 appCreator) 
{ 
  LocalID dbID;
  DmSearchStateType searchState;
  Err err;
  UInt16 cno;

  DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTPanel, appCreator, true, &cno, &dbID);
  ErrNonFatalDisplayIf(!dbID, "Could not find app");
  if(dbID)
  {
    err = SysUIAppSwitch(cno, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
    ErrNonFatalDisplayIf(err, "Could not launch app");
  }
}

void setConnectionList()
{
  FormType* frmP;
  ListType *lstP;
  UInt32 activeConnectionID;
  int i;

  frmP = FrmGetActiveForm();
  lstP = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConnectionsList));
  activeConnectionID = GetActiveService();

  for(i = 0; i < totConnections; i++)
    if(connectionsID[i] == activeConnectionID)
      LstSetSelection(lstP, i);

  if(connData.ifUp == 0)
    hideNetInfo();
  else
    displayNetInfo();
}

static Err StartApplication() 
{
  UInt16 nAppPrefsSize;
  UInt8 netStatus;

  nAppPrefsSize = sizeof(UInt8);
  if(PrefGetAppPreferences(appID, 0, &netStatus, &nAppPrefsSize, true) == noPreferenceFound)
    netStatus = 0;
  if(SysLibFind("Net.lib", &AppNetRefnum) == 0)
  {
    connected = netStatus;
    if(netStatus == 1)
      NetLibOpen(AppNetRefnum, &Instance);
  }

  FrmGotoForm(MainForm);
  return 0;
}

static void StopApplication() 
{
  int i;

  MemHandleUnlock(serviceList.IDListH);
  for(i = 0; i < serviceList.numServices; i++)
    MemPtrFree(itemList[i]);
  MemPtrFree(itemList);
  MemPtrFree(connectionsID);
  if(connected == 1)
    NetLibClose(AppNetRefnum, false);
  FrmCloseAllForms();
}

Char* GetField(UInt16 formID, UInt16 fieldID)
{
  FormPtr frm;
  FieldPtr fld;
  Char *p;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  p = FldGetTextPtr(fld);

  return (p ? p : "");
}

void SetField(UInt16 formID, UInt16 fieldID, MemPtr str)
{
  FormPtr frm;
  FieldPtr fld;
  CharPtr p;
  VoidHand h;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  h = (VoidHand) FldGetTextHandle(fld);
  if(h == NULL)
  {
    h = MemHandleNew(FldGetMaxChars(fld)+1);
    ErrFatalDisplayIf(!h, "No Memory");
  }

  p = (CharPtr) MemHandleLock(h);
  StrCopy(p, str);
  MemHandleUnlock(h);
  FldSetTextHandle(fld, (Handle) h);
}

void doWhois()
{
  Char domain[255], server[50], *answer, str[255];
  FormPtr frm = FrmGetFormPtr(WhoisForm);
  FieldAttrType attrText;
  NetSocketRef socket;
  Err error = errNone;
  NetHostInfoPtr host;
  NetHostInfoBufPtr buf;
  NetSocketAddrINType sockaddr;
  ULong **addrp;
  UInt16 nBytes;
  Word scrollPos, textHeight, fieldHeight;
  Int32 AppNetTimeout;

  AppNetTimeout = SysTicksPerSecond() * 15; // TimeOut: 15 seconds
  FldGetAttributes(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer)), &attrText);
  attrText.singleLine = 0;
  attrText.editable = 0;
  StrCopy(domain, GetField(WhoisForm, idDomainName));
  getServerForDomain(domain, server);
  answer = MemPtrNew(10240 * sizeof(Char));
  MemSet(answer, 10240, 0);

  if(StrCompare(server, "") != 0)
  {
    StrPrintF(str, "Connecting to '%s'...", server);
    SetField(WhoisForm, idStatus, str);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
    socket = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, AppNetTimeout, &error);
    if(!error)
    {
      buf = (NetHostInfoBufPtr)MemPtrNew(sizeof(NetHostInfoBufType));
      host = NetLibGetHostByName(AppNetRefnum, server, buf, AppNetTimeout, &error);
      sockaddr.family = AF_INET;
      sockaddr.port = htons(43);
      sockaddr.addr = INADDR_ANY;
      for(addrp = (ULong **)buf->hostInfo.addrListP; *addrp; ++addrp)
      {
        ULong a = htonl(**addrp);
        if(a != INADDR_ANY)
        {
          sockaddr.addr = a;
          break;
        }
      }
      MemPtrFree(buf);
      NetLibSocketConnect(AppNetRefnum, socket, (NetSocketAddrType*)&sockaddr, sizeof(NetSocketAddrType), AppNetTimeout, &error);
      StrPrintF(str, "Quering '%s'...", server);
      SetField(WhoisForm, idStatus, str);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
      NetLibSend(AppNetRefnum, socket, domain, StrLen(domain), 0x0, NULL, 0, AppNetTimeout, &error);
      NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
      nBytes = 1;
      while(nBytes != 0)
      {
        MemSet(str, 255, 0);
        nBytes = NetLibReceive(AppNetRefnum, socket, str, 250, 0x0, NULL, NULL, AppNetTimeout, &error);
        if(nBytes != 0)
          StrCat(answer, str);
      }
      SetField(WhoisForm, idStatus, "");
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
      NetLibSocketClose(AppNetRefnum, socket, AppNetTimeout, &error);
    }
    else
    {
      StrPrintF(str, "Unable to communicate with '%s'...", server);
      SetField(WhoisForm, idStatus, str);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
    }
  }
  else
    StrCopy(answer, "Invalid TLD");

  SetField(WhoisForm, idAnswer, answer);
  FldSetAttributes(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer)), &attrText);

  FldGetScrollValues(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer)), &scrollPos, &textHeight, &fieldHeight);
  if(textHeight > fieldHeight)
    SclSetScrollBar(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswerScrollbar)), scrollPos, 0, textHeight - fieldHeight, fieldHeight);
  else
    SclSetScrollBar(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswerScrollbar)), scrollPos, 0, 0, 0);
  FldSetScrollPosition(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer)), 0);

  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer)));
  MemPtrFree(answer);
}

Char encode(Char u)
{
  if(u < 26)  return 'A' + u;
  if(u < 52)  return 'a' + (u - 26);
  if(u < 62)  return '0' + (u - 52);
  if(u == 62) return '+';
  
  return '/';
}

Char *encode_base64(Char *src)
{
  int i, p, size;
  Char *out;

  if(!src)
    return NULL;

  size = StrLen(src);

  out = MemPtrNew((size * 4 / 3 +4) * sizeof(Char));
  MemSet(out, (size * 4 / 3 + 4), 0);
  p = 0;

  for(i = 0; i < size; i += 3)
  {
    Char b1=0, b2=0, b3=0, b4=0, b5=0, b6=0, b7=0;

    b1 = src[i];

    if(i + 1 < size)
      b2 = src[i + 1];

    if(i + 2 < size)
      b3 = src[i + 2];

    b4 = b1>>2;
    b5 = ((b1 & 0x3) << 4) | (b2 >> 4);
    b6 = ((b2 & 0xf) << 2) | (b3 >> 6);
    b7 = b3 & 0x3f;

    out[p++] = encode(b4);
    out[p++] = encode(b5);

    if(i + 1 < size)
      out[p++] = encode(b6);
    else
      out[p++] = '=';

    if(i + 2 < size)
      out[p++] = encode(b7);
    else
      out[p++] = '=';
  }

  return out;
}

void doDynDNS()
{
  Char *answer, str[255], dynDnsStatus[255], *userPass;
  Char username[255], password[255], hostname[255];
  FormPtr frm = FrmGetFormPtr(DynDNSForm);
  NetSocketRef socket;
  Err error = errNone;
  NetHostInfoPtr host;
  NetHostInfoBufPtr buf;
  NetSocketAddrINType sockaddr;
  ULong **addrp;
  UInt16 nBytes, l, i;
  Boolean hasToGet, commProb;
  Int32 AppNetTimeout;

  commProb = false;
  AppNetTimeout = SysTicksPerSecond() * 15; // TimeOut: 15 seconds
  StrCopy(username, GetField(DynDNSForm, idUsername));
  StrCopy(password, GetField(DynDNSForm, idPassword));
  StrCopy(hostname, GetField(DynDNSForm, idHostname));
  StrPrintF(str, "%s:%s", username, password);
  userPass = encode_base64(str);
  answer = MemPtrNew(1024 * sizeof(Char));
  MemSet(answer, 1024, 0);
  StrCopy(str, "Connecting to dyndns...");
  SetField(DynDNSForm, idStatus, str);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

  socket = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, AppNetTimeout, &error);
  if(!error)
  {
    buf = (NetHostInfoBufPtr)MemPtrNew(sizeof(NetHostInfoBufType));
    host = NetLibGetHostByName(AppNetRefnum, "members.dyndns.org", buf, AppNetTimeout, &error);
    if(host != 0)
    {
      sockaddr.family = AF_INET;
      sockaddr.port = htons(80);
      sockaddr.addr = INADDR_ANY;
      for(addrp = (ULong **)buf->hostInfo.addrListP; *addrp; ++addrp)
      {
        ULong a = htonl(**addrp);
        if(a != INADDR_ANY)
        {
          sockaddr.addr = a;
          break;
        }
      }
      MemPtrFree(buf);
      if(NetLibSocketConnect(AppNetRefnum, socket, (NetSocketAddrType*)&sockaddr, sizeof(NetSocketAddrType), AppNetTimeout, &error) != -1)
      {
        StrCopy(str, "Sending host information...");
        SetField(DynDNSForm, idStatus, str);
        FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

        StrPrintF(str, "GET /nic/update?system=dyndns&hostname=%s HTTP/1.0", hostname);
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
        StrCopy(str, "User-Agent: NetTools 1.1");
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
        StrCopy(str, "Host: members.dyndns.org");
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
        StrCopy(str, "Accept: */*");
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
        StrCopy(str, "Connection: Keep-Alive");
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
        StrPrintF(str, "Authorization: Basic %s", userPass);
        NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
        NetLibSend(AppNetRefnum, socket, "\r\n\r\n", 4, 0x0, NULL, 0, AppNetTimeout, &error);
        nBytes = 1;
        while(nBytes != 0)
        {
          MemSet(str, 255, 0);
          nBytes = NetLibReceive(AppNetRefnum, socket, str, 250, 0x0, NULL, NULL, AppNetTimeout, &error);
          if(nBytes != 0)
            StrCat(answer, str);
        }
        hasToGet = false;
        MemSet(str, 255, 0);
        MemSet(dynDnsStatus, 255, 0);
        for(i = l = 0; l < StrLen(answer); l++)
        {
          if(answer[l] == '\n')
          {
            l++;
            i = 0;
            MemSet(str, 255, 0);
          }
          if(answer[l] != '\r' && answer[l] != '\n')
            str[i++] = answer[l];
          else
          {
            if(hasToGet == true && StrCompare(str, "") != 0)
              StrPrintF(dynDnsStatus, "DynDNS says: [%s]", str);
            if(StrCompare(str, "Connection: close") == 0)
              hasToGet = true;
          }
        }
        if(hasToGet == true && StrCompare(str, "") != 0)
          StrPrintF(dynDnsStatus, "DynDNS: [%s]", str);
        SetField(DynDNSForm, idStatus, dynDnsStatus);
        FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

        NetLibSocketClose(AppNetRefnum, socket, AppNetTimeout, &error);
      }
      else
        commProb = true;
    }
    else
      commProb = true;
  }
  else
    commProb = true;

  if(commProb == true)
  {
    StrCopy(str, "Unable to communicate with dyndns!");
    SetField(DynDNSForm, idStatus, str);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
  }

  MemPtrFree(answer);
  MemPtrFree(userPass);
}

Boolean WhoisFormEventHandler(EventPtr event)
{
  Boolean handled = false;
  FormPtr frm = FrmGetFormPtr(WhoisForm);
  FieldPtr fld = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idAnswer));
  Word scrollPos, textHeight, fieldHeight;

  switch(event->eType)
  {
    case frmOpenEvent:
      FrmDrawForm(FrmGetFormPtr(WhoisForm));
      if(connectedToNet == 1)
        CtlSetEnabled(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idButtonExecute)), true);
      else
        CtlSetEnabled(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idButtonExecute)), false);
      handled = true;
    break;

    case frmCloseEvent:
      FrmEraseForm(FrmGetFormPtr(WhoisForm));
      FrmDeleteForm(FrmGetFormPtr(WhoisForm));
      handled = true;
    break;

    case sclRepeatEvent:
    {
      FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);
      if (scrollPos < event->data.sclExit.newValue)
        FldScrollField(fld, event->data.sclExit.newValue - scrollPos, winDown);
      else
        FldScrollField(fld, scrollPos - event->data.sclExit.newValue, winUp);
      FldDrawField(fld);
      handled = true;
      break;
    }

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID)
      {
        case idButtonExecute:
          doWhois();
          handled = true;
        break;
        case idButtonCancel:
          FrmReturnToForm(MainForm);
          handled = true;
        break;
      }
    break;

    default:
    break;
  }

  return handled;
}

Boolean DynDNSFormEventHandler(EventPtr event)
{
  Boolean handled = false;
  FormPtr frm = FrmGetFormPtr(DynDNSForm);
  FieldPtr fld;
  UInt16 nAppPrefsSize;

  switch(event->eType)
  {
    case frmOpenEvent:
      FrmDrawForm(FrmGetFormPtr(DynDNSForm));
      nAppPrefsSize = sizeof(Preferences);
      if(PrefGetAppPreferences(appID, 1, &gAppPrefs, &nAppPrefsSize, true) == noPreferenceFound)
        MemSet(&gAppPrefs, sizeof(Preferences), 0);
      CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)), gAppPrefs.hasToDisplayPassword);
      if(CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword))))
        FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), stdFont);
      else
        FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), symbol11Font);
      SetField(DynDNSForm, idUsername, gAppPrefs.username);
      SetField(DynDNSForm, idPassword, gAppPrefs.password);
      SetField(DynDNSForm, idHostname, gAppPrefs.hostname);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idUsername)));
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)));
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idHostname)));
      if(connectedToNet == 1)
        CtlSetEnabled(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idButtonExecute)), true);
      else
        CtlSetEnabled(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idButtonExecute)), false);
      handled = true;
    break;

    case frmCloseEvent:
      FrmEraseForm(FrmGetFormPtr(DynDNSForm));
      FrmDeleteForm(FrmGetFormPtr(DynDNSForm));
      handled = true;
    break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID)
      {
        case displayPassword:
          if(CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword))))
            FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), stdFont);
          else
            FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), symbol11Font);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)));
          handled = true;
        break;
        case idButtonExecute:
          StrCopy(gAppPrefs.username, GetField(DynDNSForm, idUsername));
          StrCopy(gAppPrefs.password, GetField(DynDNSForm, idPassword));
          StrCopy(gAppPrefs.hostname, GetField(DynDNSForm, idHostname));
          gAppPrefs.hasToDisplayPassword = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)));
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
          doDynDNS();
          handled = true;
        break;
        case idButtonCancel:
          StrCopy(gAppPrefs.username, GetField(DynDNSForm, idUsername));
          StrCopy(gAppPrefs.password, GetField(DynDNSForm, idPassword));
          StrCopy(gAppPrefs.hostname, GetField(DynDNSForm, idHostname));
          gAppPrefs.hasToDisplayPassword = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)));
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
          FrmReturnToForm(MainForm);
          handled = true;
        break;
      }
    break;

    default:
    break;
  }

  return handled;
}

Boolean FormEventHandler(EventPtr event) 
{
  Boolean handled = false;

  switch(event->eType)
  {
    case frmOpenEvent:
      gpForm = FrmGetFormPtr(MainForm);
      FrmDrawForm(gpForm);
      handled = true;
    break;

    case frmCloseEvent:
      FrmEraseForm(gpForm);
      FrmDeleteForm(gpForm);
      gpForm = 0;
      handled = true;
    break;

    default:
    break;
  }

  return handled;
}

static Boolean ApplicationEventHandler(EventPtr event) 
{
  Boolean handled = false;
  FormType* frmP;
  ListType *lstP;

  switch(event->eType)
  {
    case frmLoadEvent:
    {
      FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);
      FrmSetActiveForm(pForm);
      switch(event->data.frmLoad.formID)
      {
        case MainForm:
          FrmSetEventHandler(pForm, FormEventHandler);
          initConnectionsList();
          getConnectionData();
          setConnectionList();
        break;

        case WhoisForm:
          FrmSetEventHandler(pForm, WhoisFormEventHandler);
        break;

        case DynDNSForm:
          FrmSetEventHandler(pForm, DynDNSFormEventHandler);
        break;

        default:
        break;
      }
      handled = true;
    }
    break;

    case menuEvent:
      switch(event->data.menu.itemID)
      {
      	case HelpMenuAbout:
          FrmAlert(AboutAlert);
        break;
      	case ToolsMenuWhois:
          FrmPopupForm(WhoisForm);
        break;
      	case ToolsMenuDynDNS:
          FrmPopupForm(DynDNSForm);
        break;
      	case DataMenuRefresh:
          if(connected == 1)
          {
            getConnectionData();
            setConnectionList();
          }
        break;
      }
      handled = true;
    break;

    case ctlSelectEvent:
    {
      frmP = FrmGetActiveForm();
      switch(event->data.ctlSelect.controlID)
      {
      	case idButtonEdit:
      	{
          frmP = FrmGetActiveForm();
          lstP = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConnectionsList));
          SetActiveService(connectionsID[LstGetSelection(lstP)]);
          CallPanel(sysFileCNetworkPanel);
      	  handled = true;
      	  break;
        }
      	case idButtonConnect:
      	{
          frmP = FrmGetActiveForm();
          lstP = (ListType *) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ConnectionsList));
          SetActiveService(connectionsID[LstGetSelection(lstP)]);
          if(connectToService() == 0)
          {
            getConnectionData();
            if(connData.ifUp == 0)
              hideNetInfo();
            else
              displayNetInfo();
          }
      	  handled = true;
      	  break;
      	}
      	case idButtonDisconnect:
      	{
          disconnectToService();
          hideNetInfo();
      	  handled = true;
      	  break;
      	}
      }
    }
    break;

    default:
    break;
  }

  return handled;
}

static void EventLoop() 
{
  EventType event;
  UInt16 error;

  do
  {
    EvtGetEvent(&event, evtWaitForever);

    if (!SysHandleEvent(&event))
    {
      if (!MenuHandleEvent(0, &event, &error))
      {
        if (!ApplicationEventHandler(&event))
          FrmDispatchEvent(&event);
      }
    }
  } while(event.eType != appStopEvent);
}

static void RegisterForNotifications()
{
  UInt16 cardNo;
  LocalID dbID;
 
  if(0 != SysCurAppDatabase(&cardNo, &dbID))
    return;
 
  SysNotifyRegister(cardNo, dbID, sysNotifyLateWakeupEvent, NULL, sysNotifyNormalPriority, 0);
  SysNotifyRegister(cardNo, dbID, sysNotifyNetLibIFMediaEvent, NULL, sysNotifyNormalPriority, 0);
}

static void HandleNotification(SysNotifyParamType *np)
{
  SysNotifyNetLibIFMediaType *details;
  UInt8 netStatus;

  if(np->notifyType == sysNotifyLateWakeupEvent)
  {
    netStatus = 0;
    PrefSetAppPreferences(appID, 0, versionID, &netStatus, sizeof(netStatus), true);
  }
  if(np->notifyType == sysNotifyNetLibIFMediaEvent)
  {
    details = (SysNotifyNetLibIFMediaType*) np->notifyDetailsP;
    if(details->eType == netIFMediaUp)
    {
      netStatus = 1;
      PrefSetAppPreferences(appID, 0, versionID, &netStatus, sizeof(netStatus), true);
    }
    else
    {
      netStatus = 0;
      PrefSetAppPreferences(appID, 0, versionID, &netStatus, sizeof(netStatus), true);
    }
  }
}

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
  Err err = 0;
  UInt32 romVersion;

  switch(launchCode)
  {
    case sysAppLaunchCmdSystemReset:
    case sysAppLaunchCmdSyncNotify:
      RegisterForNotifications();
    break;
 
    case sysAppLaunchCmdNotify:
      HandleNotification((SysNotifyParamType*) cmdPBP);
    break;

    case sysAppLaunchCmdNormalLaunch:
    {
      FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
      if(romVersion < sysMakeROMVersion(3, 5, 0, sysROMStageRelease, 0))
      {
        FrmAlert(RomIncompatibleAlert);
        return sysErrRomIncompatible;
      }
      if((err = StartApplication()) == 0)
      {
        EventLoop();
        StopApplication();
      }
      break;
    }
  }

  return err;
}
